@@ -41,6 +41,7 @@
 #include <signal.h>
 #include <err.h>
 #include <errno.h>
+#include <time.h>
 
 #include "config.h"
 
@@ -56,6 +57,9 @@ ssize_t atomicio(ssize_t (*f)(), int, void *, size_t);
 void http_connectioncb(int, short, void *);
 int http_isrobotstxt(struct uri *);
 
+extern int debug;
+#define DFPRINTF(x,y)   if (debug >= x) fprintf y
+
 TAILQ_HEAD(uri_list, uri_small) uriqueue;
 TAILQ_HEAD(dispatch_list, dispatch) dispatchqueue;
 
@@ -102,24 +106,36 @@ uri_new(void)
 		return (NULL);
 
 	uri->fd = -1;
-
+	uri->save_fd = -1;
+	
 	return (uri);
 }
 
 void
 uri_free(struct uri *uri)
 {
+	if (event_initialized(&uri->ev))
+		event_del(&uri->ev);
+	
 	/* If this is an aborted robots.txt, we treat it as no robots.txt */
 	if (use_robots)
 		http_isrobotstxt(uri);
 
 	if (uri->fd != -1) {
+		int n = 0;
+
 		close(uri->fd);
 		connections--;
 
-		while (http_postevent() != -1)
-			;
+		/* Try to start new connections */
+		if (TAILQ_FIRST(&uriqueue) != NULL) {
+			while (http_postevent() != -1)
+				n++;
+		}
 	}
+	
+	if (uri->save_fd != -1)
+		close(uri->save_fd);
 
 	if (uri->uri != NULL)
 		free(uri->uri);
@@ -158,6 +174,7 @@ terminate_handler(int sig)
 	int s = errno;
 
 	signal(SIGINT, SIG_IGN);
+	signal(SIGPIPE, SIG_IGN);
 	terminate = 1;
 
 	errno = s;
@@ -189,8 +206,12 @@ http_init(char *filename)
 			perror("signal");
 			return (-1);
 		}
+		if (signal(SIGPIPE, terminate_handler) == SIG_ERR) {
+			perror("signal");
+			return (-1);
+		}
 	}
-
+	
 	return (0);
 }
 
@@ -211,10 +232,11 @@ http_print_stats(void)
 		(float)(http_stats.gets + http_stats.heads)/ spent);
 	fprintf(stdout, "Throughput %f KBytes/sec\n",
 		(float) bytes / 1024 / spent);
-	fprintf(stdout, "Max fifo size: %d, %f Kbyte\n",
+	fprintf(stdout, "Max fifo size: %d, %f Kbyte, current: %d/%d\n",
 		http_stats.maxsize,
 		(float) http_stats.maxsize *
-		(sizeof(struct uri_small) + 80)/1024);
+		(sizeof(struct uri_small) + 80)/1024,
+		http_stats.fifosize, connections);
 
 	if (use_robots)
 		http_robots_print_stats();
@@ -331,6 +353,9 @@ http_connection_retry(struct uri *uri)
 	if (uri->retry > HTTP_MAXRETRY)
 		goto fail;
 
+	/* As we are closing a file descriptor here, try to start
+	 * a new connection on failure
+	 */
 	close(uri->fd);
 	connections--;
 
@@ -343,6 +368,8 @@ http_connection_retry(struct uri *uri)
  fail:
 	warnx(__FUNCTION__": failed retry: %s", uri->uri);
 	uri_free(uri);
+
+	http_postevent();
 }
 
 void
@@ -514,8 +541,11 @@ http_postevent(void)
 	char *host = NULL;
 	u_short port;
 
-	if (connections >= http_maxconnects)
+	if (connections >= http_maxconnects) {
+		DFPRINTF(2, (stderr, __FUNCTION__": %d >= http_maxconnects\n",
+			     connections));
 		return (-1);
+	}
 
 	if (terminate) {
 		if (connections == 0) {
@@ -549,11 +579,10 @@ http_postevent(void)
 
 			if ((htrobot = http_robots_find(host, port)) == NULL) {
 				if (http_robots_get(host, port) != -1)
-					goto out;
+					continue;
 			} else {
 				if (htrobot->flags & HTROBOT_PENDING)
 					continue;
-
 				if (!http_robots_allow(htrobot, suri->uri)) {
 					fprintf(stderr, __FUNCTION__
 						": robots diallowed url: %s\n",
@@ -829,21 +858,28 @@ http_parseheader(struct uri *uri)
 
 	/* If we just had a HEAD request, we are done now */
 	if (uri->flags & HTTP_REQUEST_HEAD) {
-		http_dispatch(uri);
+		http_dispatch(uri, HTTPDIS_FREE);
 		return;
 	}
 	
 	if (uri->length != -1) {
+		size_t len;
 		u_char *p;
 
-		p = realloc(uri->body, uri->length + 1);
+		len = uri->length;
+		if (len > HTTP_MAXMEM)
+			len = HTTP_MAXMEM;
+		
+		p = realloc(uri->body, len + 1);
 		if (p == NULL) {
 			warn(__FUNCTION__": malloc");
 			uri_free(uri);
 			return;
 		}
 		uri->body = p;
-		uri->body[uri->length] = '\0';
+		uri->body[len] = '\0';
+		uri->bdmemlen = len;
+		uri->bdread = uri->bdlen;
 	}
 
 	event_set(&uri->ev, uri->fd, EV_READ,
@@ -878,8 +914,10 @@ http_readbody(int fd, short event, void *arg)
 		where = body;
 		len = sizeof(body);
 	} else {
-		where = uri->body + uri->bdlen;
+		where = uri->body + uri->bdread;
 		len = uri->length - uri->bdlen;
+		if (len > uri->bdmemlen - uri->bdread)
+			len = uri->bdmemlen - uri->bdread;
 	}
 
 	n = read(uri->fd, where, len);
@@ -918,8 +956,10 @@ http_readbody(int fd, short event, void *arg)
 		memcpy(uri->body + uri->bdlen, body, n);
 		uri->bdlen += n;
 		uri->body[uri->bdlen] = '\0';
-	} else
+	} else {
 		uri->bdlen += n;
+		uri->bdread += n;
+	}
 
 	if (uri->length == -1 || uri->bdlen < uri->length)
 		goto readmore;
@@ -927,13 +967,17 @@ http_readbody(int fd, short event, void *arg)
 	/* We are done with this document */
 
  done:
-	http_dispatch(uri);
+	http_dispatch(uri, HTTPDIS_FREE);
 	return;
 
  readmore:
 	timerclear(&tv);
 	tv.tv_sec = HTTP_READTIMEOUT;
 	event_add(&uri->ev, &tv);
+
+	if (uri->length != -1 && uri->bdmemlen == uri->bdread)
+		http_dispatch(uri, HTTPDIS_KEEP);
+	
 	return;
 }
 
@@ -957,7 +1001,7 @@ http_isrobotstxt(struct uri *uri)
 
 
 void
-http_dispatch(struct uri *uri)
+http_dispatch(struct uri *uri, int mustfree)
 {
 	struct dispatch *dispatch;
 	char *format;
@@ -988,7 +1032,13 @@ http_dispatch(struct uri *uri)
 		}
 	}
 
-	uri_free(uri);
+	/* caller decides */
+	if (mustfree || dispatch == NULL) {
+		uri_free(uri);
+		return;
+	}
+
+	/* object is still alive */
 }
 
 char *
